////////////////////////////////////////////////////////////////////////////////
// main_dn.cpp
// Implmentation for s11nconvert utility, to convert data between the various
// Serializers.
// Author: stephan@s11n.net
// License: Public Domain
////////////////////////////////////////////////////////////////////////////////


#include <iostream>
#include <string>
#include <memory> // auto_ptr
#include <unistd.h> // isatty()

#include <s11n.net/s11n/variant.hpp>
#include <s11n.net/s11n/s11n_debuggering_macros.hpp> // CERR
#include <s11n.net/s11n/s11n.hpp> // s11n core framework
#include <s11n.net/s11n/s11nlite.hpp> // s11nlite framework
#include <s11n.net/s11n/io/serializers.hpp> // utility code for s11n::io
#include <s11n.net/s11n/s11n_config.hpp>

#include <s11n.net/s11n/traits.hpp> //node_traits
#include "argv_parser.hpp"

#if s11n_HAVE_ZFSTREAM
#  include <s11n.net/zfstream/zfstream.hpp>
#endif


// DLLS_WORKIE is a temporary workaround for no DLL support in 1.1
#define DLLS_WORKIE (0)

bool verbose = false;
#define VERBOSE if( verbose ) CERR

void show_version_info( std::ostream & os )
{
        os << "s11nconvert version " <<  s11n_S11N_LIBRARY_VERSION << "\n"
	   << "linked against s11n version " << s11n::library_version() << "\n"
	   << "A tool to convert data between the formats supported by libs11n.\n"
	   << "License: " << s11n_PACKAGE_LICENSE << "\n"
	   << "This software comes with NO WARRANTY of any type.\n"
	   << "Find out more about s11n at http://s11n.net" << std::endl;
}

void show_help( std::ostream & os )
{

	show_version_info( os );

	os << "\nCommand-line options:\n\n";

#define HOP(A,H) os << "\t" << A << ":\n\t\t" << H << "\n";
	HOP("-?, --help", "Shows help text on stdout and exits with code 0." );
	HOP("-f,--file FILENAME", "Sets input file name. The name '-' means stdin." );
	HOP("-K, --known-serializers", "Listed known/loaded Serializers (format handlers).");
	HOP("-o, --output FILENAME", "Sets output filename. The name '-' means stdout." );
	HOP("-S SERIALIZER", "Forces the specified input Serializer to be used." );
	HOP("-s, --format SERIALIZER", "Sets the output Serializer (i.e., output format)." );
	HOP("-v, --verbose", "Enabled verbose mode. Additional info will go to stderr." );
	HOP("-V, --version", "Shows version info on stdout and exits with code 0." );

	HOP("-z","Compress output files with gzip, if you have libzsfstream and it is built with this support. Works only on files, not stdout." );
	HOP("-bz","As for the -z option, but applies to bzip2 compression. -z and -bz may not be used together." );
#undef HOP

	os << "\nActually, all arguments may be in any of these forms:\n"
	   << "\t-ARG VALUE  (an empty value is assumed to be a boolean 1)\n"
	   << "\t-ARG=VALUE (quote the value if it contains spaces)\n"
	   << "The number of leading dashes for an argument is absolutely irrelevant. "
	   << "When an argument is specified more than once, only the last one set takes effect.\n";

	os << "\nSeveral options from s11nconvert 1.0.x are currently not available, "
	   << "but will be re-added eventually.\n";

	os << std::endl;

}

int
main( int argc, char **argv )
{
#define ERR std::cerr << "s11nconvert error: "

	cliutil::argv_map args;
	cliutil::unnamed_args_list unused;

	typedef s11nlite::node_traits NTR;
	s11nlite::node_type argn; // we're gonna cheat a bit and use this for arg parsing

	//int ArgC = 
		cliutil::parse_args( argc, argv, 1, args, unused );

	cliutil::argv_map::const_iterator amit = args.begin();
	for( ; args.end() != amit; ++amit )
	{
		NTR::set( argn, (*amit).first, (*amit).second );
	}

        if( NTR::is_set( argn, "v" ) || NTR::is_set( argn, "verbose" ) )
        {
                verbose = true;
        }
        VERBOSE << "Verbose mode on.\n";

        enum {
        ErrorNoSerializer = 1,
        ErrorFileNotFound,
        ErrorNoInputFile,
        ErrorInputReadFailed,
        ErrorOutputOpenFailed,
        ErrorLoadNodeFailed,
        ErrorSaving,
        ErrorDLL,
        ErrorForcedSerializer,
        ErrorUsageError,
        ErrorNoSerializers
        };



        ////////////////////////////////////////////////////////////////////////
        // Load list of known/registered Serializers.
        // Do this AFTER loading DLLs so that DLLs may install additional Serializers...
        std::string known_ser;
        typedef std::list< std::string > SerList;
        SerList serlist;
        s11n::io::serializer_list<s11nlite::node_type>( serlist );
        SerList::const_iterator cit = serlist.begin(),
                cet = serlist.end();
        for( ; cet != cit; ++cit )
        {
                known_ser += (*cit) + " ";
        }
        if( known_ser.empty() )
        {
                ERR << "No data formats loaded!\n"
		    << "This means we couldn't find any built-in Serializers, "
		    << "which is probably inidicitive of a problem in your local s11n installation.\n"
		    << "Make sure that your s11nconvert is not linking against an older, internally incompatible libs11n "
		    << "by using 'ldd s11nconvert'.\n"
			;
                return ErrorNoSerializers;
        }


        if( NTR::is_set( argn, "known-serializers" ) || NTR::is_set( argn, "K" ) )
        {
                std::cout << known_ser << std::endl;
                return 0;
        }
        ////////////////////////////////////////////////////////////////////////


        if( NTR::is_set( argn, "version" ) || NTR::is_set( argn, "V" ) )
        {
                show_version_info( std::cout );
                return 0;
        }

        if( NTR::is_set( argn, "help" ) || NTR::is_set( argn, "?" ) )
        {
                show_help( std::cout );
                return 0;
        }


#if s11n_HAVE_ZFSTREAM
        if( NTR::is_set( argn, "z" ) && NTR::is_set( argn, "bz" ) )
        {
                ERR << "-z and -bz may not be used together." << std::endl;
                return ErrorUsageError;
        }
#endif // s11n_HAVE_ZFSTREAM


        std::string fmt = NTR::get( argn, "s", NTR::get( argn, "format", std::string() ) );
        if( fmt.empty() )
        {
                ERR << "No output format (serializer) specified. "
                    << "Try using [-s or --format] with one of:\n"
                    << known_ser << "\n";
                return ErrorNoSerializer;
        }


         typedef std::auto_ptr<s11nlite::serializer_interface> APSER;
         APSER ser = APSER( s11nlite::create_serializer( fmt ) );
         if( ! ser.get() )
         {
                 ERR << "No Serializer found for name '"<<fmt<<"'. Try one of:\n"
                     << known_ser << "\n";
                 return ErrorNoSerializer;
         }

 
         bool usegz = false;
         bool usebz = false;
         if( NTR::is_set( argn, "z" ) )
         {
                 usegz = true;
                 
         }
	 else if( NTR::is_set( argn, "bz" ) )
	 {
                 usebz = true;
		 
	 }

	 if( usegz || usebz )
	 {
#if s11n_HAVE_ZFSTREAM
		 zfstream::compression_policy( usegz ? zfstream::GZipCompression : zfstream::BZipCompression );
#else
		 ERR << "zlib/bzip2 compression is not supported by your libs11n.\n";
		 return ErrorUsageError;
#endif
	 }


	 std::string ifname = NTR::get( argn, "f", NTR::get( argn, "file", std::string() ) );
         std::istream * is = 0;
         bool use_infile = false;
         if( ( "-" == ifname ) || ! isatty(STDIN_FILENO) )
         {
                 is = &std::cin;
         }
         else if( ifname.empty() )
         {
                 ERR << "No input file specified.\n";
                 return ErrorNoInputFile;
         }
         else
         {
                 use_infile = true;
         }



         std::ostream * os = 0;

	 std::string ofname = NTR::get( argn, "o", NTR::get(argn, "output", std::string() ) );
         bool use_ofile = false;
         if( ofname.empty() || ("-" == ofname) )
         {
                 if( usegz || usebz )
                 {
                         ERR << "Compression (-z and -bz) works only for files, not cout." << std::endl;
                         return ErrorUsageError;
                 }
                 os = &std::cout;
         }
         else
         {
                 use_ofile = true;
         }

         typedef std::auto_ptr<s11nlite::node_type> NAP;
         NAP innode;
         std::string forcedinser = NTR::get( argn, "S", std::string() );
         if( ! forcedinser.empty() )
         {
                 VERBOSE << "Forcing input serializer: " << forcedinser << "\n";
                 s11nlite::serializer_interface * forceser = s11nlite::create_serializer( forcedinser );
                 if( 0 == forceser )
                 {
                         ERR << "Could not load forced input serializer '"<<forcedinser<<"'!\n";
                         return ErrorForcedSerializer;
                 }
                 innode = NAP( use_infile
                               ? forceser->deserialize( ifname )
                               : forceser->deserialize( *is ) );
         }
         else
         {
                 innode = NAP( use_infile ? s11nlite::load_node( ifname ) : s11nlite::load_node( *is ) );
         }
         if( ! innode.get() )
         {
                 ERR << "Error loading node tree from input stream!\n";
                 return ErrorLoadNodeFailed;
         }

        // And FINALLY... we get down to the real work...
        // these couple lines almost now don't seem worth all the hassle ;)

	 bool workie = use_ofile ? ser->serialize( *innode, ofname ) : ser->serialize( *innode, *os );
	 if( ! workie )
	 {
		 ERR << "Error saving output!\n";
		 return ErrorSaving;
	 }

	return 0;
#undef ERR
};



